;;;; Copyright © 2016 Federico Beffa <beffa@fbengineering.ch>
;;;;
;;;; This library is free software; you can redistribute it and/or
;;;; modify it under the terms of the GNU Lesser General Public
;;;; License as published by the Free Software Foundation; either
;;;; version 3 of the License, or (at your option) any later version.
;;;; 
;;;; This library is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;;; Lesser General Public License for more details.
;;;; 
;;;; You should have received a copy of the GNU Lesser General Public
;;;; License along with this library; if not, write to the Free Software
;;;; Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

;;; Code

(define-module (restart restart-rnrs)
  #:use-module ((srfi srfi-1))
  #:use-module ((srfi srfi-26))
  #:use-module ((ice-9 match))
  #:use-module ((system repl command))
  #:use-module ((system repl common))
  #:use-module ((system repl debug))
  #:use-module ((system vm frame))
  #:export (restart))

;; from (system vm program)
(define (binding:name b) (list-ref b 0))
(define (binding:boxed? b) (list-ref b 1))
(define (binding:index b) (list-ref b 2))
(define (binding:start b) (list-ref b 3))
(define (binding:end b) (list-ref b 4))

(define raise-object-wrapper?
  (@@ (rnrs records procedural) raise-object-wrapper?))

(define raise-object-wrapper-obj
  (@@ (rnrs records procedural) raise-object-wrapper-obj))

(define-meta-command ((restart restart) repl . form)
  "restart [INDEX]
List the available restarts or call INDEX."
  (let* ((stack (make-stack #t))
         (frame (stack-ref stack 8))
         (bindings (frame-bindings frame))
         (restarts (frame-restarts frame))
         (restarts-alist (map cons
                              (iota (length restarts) 1)
                              restarts))
         (index (match form
                  (() 0)
                  (((? number? no)) no)
                  (_ -1))))
    (cond
     ((zero? index)
      (format #t "The following restarts are available:~%")
      (print-condition-restarters restarts-alist))
     ((and (< 0 index) (>= (length restarts) index))
      (restart-interactively (assoc-ref restarts-alist index)))
     (else
      (format #t "Restart index out of range or not a number.~%")))))

(define (frame-restarts frame)
  (apply append
         (filter-map (cut binding-is-restartable-condition? frame <>)
                     (frame-bindings frame))))

(define (binding-is-restartable-condition? frame binding)
  ;; The following let binding was copied from 'print-locals' in
  ;; (system repl debug).
  (let ((v (let ((x (frame-local-ref frame (binding:index binding))))
             (if (binding:boxed? binding)
                 (variable-ref x)
                 x))))
    (if (and (pair? v) (raise-object-wrapper? (car  v)))
        (let ((condition (raise-object-wrapper-obj (car v))))
          (if (restartable-condition? condition)
              (condition-restarters condition)
              #f))
        #f)))

(define (print-condition-restarters restarts-alist)
  (if (null? restarts-alist)
      (format #t "No restarts.~%")
      (for-each
       (match-lambda
         ((number . restart)
          (format #t "~2,T~S:~5,T(~a)~25,T~a~%"
                  number
                  (restarter-tag restart)
                  (restarter-description restart))))
       restarts-alist)))
